package query;

import global.Minibase;
import global.RID;
import global.SearchKey;
import heap.HeapFile;
import index.HashIndex;
import parser.AST_Delete;
import relop.FileScan;
import relop.Predicate;
import relop.Schema;
import relop.Tuple;

/**
 * Execution plan for deleting tuples.
 */
class Delete implements Plan {

	/** Name of the table to delete from. */
	protected String		fileName;

	private Schema			schema;

	/** Predicates to select tuples to delete. */
	protected Predicate[][]	predicates;

	/**
	 * Optimizes the plan, given the parsed query.
	 * 
	 * @throws QueryException
	 *             if table doesn't exist or predicates are invalid
	 */
	public Delete(AST_Delete tree) throws QueryException {
		// Validate that table exists
		fileName = tree.getFileName();
		schema = QueryCheck.tableExists(fileName);

		// Validate predicates
		predicates = tree.getPredicates();
		QueryCheck.predicates(schema, predicates);

	} // public Delete(AST_Delete tree) throws QueryException

	/**
	 * Executes the plan and prints applicable output.
	 */
	public void execute() {
		HeapFile tFile = new HeapFile(fileName);
		FileScan scan = new FileScan(schema, tFile);
		IndexDesc[] indices = Minibase.SystemCatalog.getIndexes(fileName);
		HashIndex ixUpdate;
		SearchKey ixKey;

		int rowCount = 0;

		boolean pass = false;
		Tuple t;
		while(scan.hasNext())
		{
			t = scan.getNext();
			//Evaluate tuple on predicates - outer loop is AND on a clause
			//Inner loop is OR
			for(Predicate[] andPreds : predicates)
			{
				pass = false;
				for(Predicate orPreds : andPreds)
				{
					if(orPreds.evaluate(t))
					{
						pass = true;
						break;
					}
				}
				//If any full predicate clause evaluates to false, we don't qualify
				if(!pass) break;
			}
			//If we qualify, delete record from table and all associated indices
			if(pass)
			{
				rowCount++;
				
				RID r = scan.getLastRID();
				
				//Remove entry from indices first, as we need the SearchKey
				for(IndexDesc ix : indices)
				{
					ixUpdate = new HashIndex(ix.indexName);
					ixKey = new SearchKey(t.getField(ix.columnName));
					ixUpdate.deleteEntry(ixKey, r);
				}
				//Now we can remove the tuple
				tFile.deleteRecord(r);
			}
		}
		scan.close();
		
		// Update system catalog for relation
		RID fRID = Minibase.SystemCatalog.getFileRID(fileName, true);
		Tuple rel_cat = new Tuple(Minibase.SystemCatalog.s_rel,
				Minibase.SystemCatalog.f_rel.selectRecord(fRID));
		rel_cat.setField(1, rel_cat.getIntFld(1) - rowCount);
		Minibase.SystemCatalog.f_rel.updateRecord(fRID, rel_cat.getData());
		// print the output message
		System.out.println("DELETE from table (" + fileName + "): " + rowCount + " rows affected.");

	} // public void execute()

} // class Delete implements Plan
