package query;

import global.Minibase;
import global.RID;
import global.SearchKey;
import heap.HeapFile;
import index.HashIndex;
import parser.AST_Update;
import relop.FileScan;
import relop.Predicate;
import relop.Schema;
import relop.Tuple;

/**
 * Execution plan for updating tuples.
 */
class Update implements Plan {

	protected String		fileName;

	protected Schema		schema;

	protected int[]			updateFields;

	protected Object[]		updateValues;

	protected Predicate[][]	updatePredicates;

	/**
	 * Optimizes the plan, given the parsed query.
	 * 
	 * @throws QueryException
	 *             if invalid column names, values, or predicates
	 */
	public Update(AST_Update tree) throws QueryException {
		// Validate table name
		fileName = tree.getFileName();
		schema = QueryCheck.tableExists(fileName);

		// Validate columns in table to update
		updateFields = QueryCheck.updateFields(schema, tree.getColumns());

		// Validate values
		updateValues = tree.getValues();
		QueryCheck.updateValues(schema, updateFields, updateValues);

		// Validate predicates
		updatePredicates = tree.getPredicates();
		QueryCheck.predicates(schema, updatePredicates);
	} // public Update(AST_Update tree) throws QueryException

	/**
	 * Executes the plan and prints applicable output.
	 */
	public void execute() {
		HeapFile tFile = new HeapFile(fileName);
		FileScan scan = new FileScan(schema, tFile);
		IndexDesc[] indices = Minibase.SystemCatalog.getIndexes(fileName, schema, updateFields);
		HashIndex ixUpdate;
		SearchKey ixKey;
		
		int rowCount = 0;
		
		boolean pass = false;
		Tuple t;
		while(scan.hasNext())
		{
			t = scan.getNext();
			//Evaluate tuple on predicates - outer loop is AND on a clause
			//Inner loop is OR
			for(Predicate[] andPreds : updatePredicates)
			{
				pass = false;
				for(Predicate orPreds : andPreds)
				{
					if(orPreds.evaluate(t))
					{
						pass = true;
						break;
					}
				}
				//If any full predicate clause evaluates to false, we don't qualify
				if(!pass) break;
			}
			//If we qualify, update record in table and all associated indices
			if(pass)
			{
				rowCount++;

				RID r = scan.getLastRID();
				
				//Used for finding SearchKey of old value in indices
				Tuple tCopy = new Tuple(schema, t.getData().clone());
				//Update tuple
				for(int i = 0; i < updateFields.length; i++)
					t.setField(updateFields[i], updateValues[i]);

				tFile.updateRecord(r, t.getData());
				//Update entry in indices
				for(IndexDesc ix : indices)
				{
					ixUpdate = new HashIndex(ix.indexName);
					//Delete old entry
					ixKey = new SearchKey(tCopy.getField(ix.columnName));
					ixUpdate.deleteEntry(ixKey, r);
					//Insert new entry
					ixKey = new SearchKey(t.getField(ix.columnName));
					ixUpdate.insertEntry(ixKey, r);
				}
			}
			
		}

		scan.close();
		// print the output message
		System.out.println("UPDATE table (" + fileName + "): " + rowCount + " rows affected.");

	} // public void execute()

} // class Update implements Plan
